from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from apify_client import ApifyClient, ApifyClientAsync
from langchain_core.utils import get_from_dict_or_env
from pydantic import BaseModel, ConfigDict, model_validator

from langchain_apify.document_loaders import ApifyDatasetLoader
from langchain_apify.utils import create_apify_client

if TYPE_CHECKING:
    from langchain_core.documents import Document


class ApifyWrapper(BaseModel):
    """Wrapper around Apify client for LangChain.

    To use, you should have the environment variable `APIFY_API_TOKEN` set
    with your API key, or pass `apify_api_token`
    as a named parameter to the constructor.

    For details, see https://docs.apify.com/platform/integrations/langchain

    Example:
        .. code-block:: python

            from langchain_apify import ApifyWrapper
            from langchain_core.documents import Document

            apify = ApifyWrapper()

            loader = apify.call_actor(
                actor_id="apify/website-content-crawler",
                run_input={
                    "startUrls": [
                        {"url": "https://python.langchain.com/docs/introduction/"}
                    ],
                    "maxCrawlPages": 10,
                    "crawlerType": "cheerio"
                },
                dataset_mapping_function=lambda item: Document(
                    page_content=item["text"] or "",
                    metadata={"source": item["url"]}
                ),
            )
            documents = loader.load()
    """

    # allow arbitrary types in the model config for the apify client fields
    model_config = ConfigDict(arbitrary_types_allowed=True)

    apify_client: ApifyClient
    apify_client_async: ApifyClientAsync
    apify_api_token: str | None = None

    def __init__(
        self,
        apify_api_token: str | None = None,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize the loader with an Apify dataset ID and a mapping function.

        Args:
            dataset_id (str): The ID of the dataset on the Apify platform.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to an instance
                of the Document class.
            apify_api_token (Optional[str]): Apify API token.
            *args: Any: Additional positional arguments.
            **kwargs: Any: Additional keyword arguments.
        """
        kwargs.update({'apify_api_token': apify_api_token})
        super().__init__(*args, **kwargs)

    @model_validator(mode='before')
    @classmethod
    def validate_environment(cls, values: dict) -> Any:  # noqa: ANN401
        """Validate environment.

        Validate that an Apify API token is set and the apify-client
        Python package exists in the current environment.

        Returns:
            Any: The validated values.
        """
        apify_api_token = get_from_dict_or_env(values, 'apify_api_token', 'APIFY_API_TOKEN')

        values['apify_client'] = create_apify_client(ApifyClient, apify_api_token)
        values['apify_client_async'] = create_apify_client(ApifyClientAsync, apify_api_token)

        return values

    def call_actor(  # noqa: PLR0913
        self,
        actor_id: str,
        run_input: dict,
        dataset_mapping_function: Callable[[dict], Document],
        *,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout_secs: int | None = None,
    ) -> ApifyDatasetLoader:
        """Run an Actor on the Apify platform and wait for results to be ready.

        Args:
            actor_id (str): The ID or name of the Actor on the Apify platform.
            run_input (Dict): The input object of the Actor that you're trying to run.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to an
                instance of the Document class.
            build (str, optional): Optionally specifies the actor build to run.
                It can be either a build tag or build number.
            memory_mbytes (int, optional): Optional memory limit for the run,
                in megabytes.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.

        Returns:
            ApifyDatasetLoader: A loader that will fetch the records from the
                Actor run's default dataset.

        Raises:
            RuntimeError: If the Actor call fails.
        """
        if (
            actor_call := self.apify_client.actor(actor_id).call(
                run_input=run_input,
                build=build,
                memory_mbytes=memory_mbytes,
                timeout_secs=timeout_secs,
            )
        ) is None:
            msg = f'Failed to call Actor {actor_id}.'
            raise RuntimeError(msg)

        return ApifyDatasetLoader(
            dataset_id=actor_call['defaultDatasetId'],
            dataset_mapping_function=dataset_mapping_function,
        )

    async def acall_actor(  # noqa: PLR0913
        self,
        actor_id: str,
        run_input: dict,
        dataset_mapping_function: Callable[[dict], Document],
        *,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout_secs: int | None = None,
    ) -> ApifyDatasetLoader:
        """Run an Actor on the Apify platform and wait for results to be ready.

        Args:
            actor_id (str): The ID or name of the Actor on the Apify platform.
            run_input (Dict): The input object of the Actor that you're trying to run.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to
                an instance of the Document class.
            build (str, optional): Optionally specifies the actor build to run.
                It can be either a build tag or build number.
            memory_mbytes (int, optional): Optional memory limit for the run,
                in megabytes.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.

        Returns:
            ApifyDatasetLoader: A loader that will fetch the records from the
                Actor run's default dataset.

        Raises:
            RuntimeError: If the Actor call fails.
        """
        if (
            actor_call := await self.apify_client_async.actor(actor_id).call(
                run_input=run_input,
                build=build,
                memory_mbytes=memory_mbytes,
                timeout_secs=timeout_secs,
            )
        ) is None:
            msg = f'Failed to call Actor {actor_id}.'
            raise RuntimeError(msg)

        return ApifyDatasetLoader(
            dataset_id=actor_call['defaultDatasetId'],
            dataset_mapping_function=dataset_mapping_function,
        )

    def call_actor_task(  # noqa: PLR0913
        self,
        task_id: str,
        task_input: dict,
        dataset_mapping_function: Callable[[dict], Document],
        *,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout_secs: int | None = None,
    ) -> ApifyDatasetLoader:
        """Run a saved Actor task on Apify and wait for results to be ready.

        Args:
            task_id (str): The ID or name of the task on the Apify platform.
            task_input (Dict): The input object of the task that you're trying to run.
                Overrides the task's saved input.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to an
                instance of the Document class.
            build (str, optional): Optionally specifies the actor build to run.
                It can be either a build tag or build number.
            memory_mbytes (int, optional): Optional memory limit for the run,
                in megabytes.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.

        Returns:
            ApifyDatasetLoader: A loader that will fetch the records from the
                task run's default dataset.

        Raises:
            RuntimeError: If the task call fails.
        """
        if (
            task_call := self.apify_client.task(task_id).call(
                task_input=task_input,
                build=build,
                memory_mbytes=memory_mbytes,
                timeout_secs=timeout_secs,
            )
        ) is None:
            msg = f'Failed to call task {task_id}.'
            raise RuntimeError(msg)

        return ApifyDatasetLoader(
            dataset_id=task_call['defaultDatasetId'],
            dataset_mapping_function=dataset_mapping_function,
        )

    async def acall_actor_task(  # noqa: PLR0913
        self,
        task_id: str,
        task_input: dict,
        dataset_mapping_function: Callable[[dict], Document],
        *,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout_secs: int | None = None,
    ) -> ApifyDatasetLoader:
        """Run a saved Actor task on Apify and wait for results to be ready.

        Args:
            task_id (str): The ID or name of the task on the Apify platform.
            task_input (Dict): The input object of the task that you're trying to run.
                Overrides the task's saved input.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to an
                instance of the Document class.
            build (str, optional): Optionally specifies the actor build to run.
                It can be either a build tag or build number.
            memory_mbytes (int, optional): Optional memory limit for the run,
                in megabytes.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.

        Returns:
            ApifyDatasetLoader: A loader that will fetch the records from the
                task run's default dataset.

        Raises:
            RuntimeError: If the task call fails.
        """
        if (
            task_call := await self.apify_client_async.task(task_id).call(
                task_input=task_input,
                build=build,
                memory_mbytes=memory_mbytes,
                timeout_secs=timeout_secs,
            )
        ) is None:
            msg = f'Failed to call task {task_id}.'
            raise RuntimeError(msg)

        return ApifyDatasetLoader(
            dataset_id=task_call['defaultDatasetId'],
            dataset_mapping_function=dataset_mapping_function,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from apify_client import ApifyClient
from langchain_core.document_loaders.base import BaseLoader
from langchain_core.documents import Document  # noqa: TCH002
from langchain_core.utils import get_from_dict_or_env
from pydantic import BaseModel, ConfigDict, model_validator

from langchain_apify.utils import create_apify_client

if TYPE_CHECKING:
    from collections.abc import Iterator


class ApifyDatasetLoader(BaseLoader, BaseModel):
    """Load datasets from Apify web scraping, crawling, and data extraction platform.

    To use, you should have the environment variable `APIFY_API_TOKEN` set
    with your API key, or pass `apify_api_token`
    as a named parameter to the constructor.

    For details, see https://docs.apify.com/platform/integrations/langchain

    Example:
        .. code-block:: python

            from langchain_apify import ApifyDatasetLoader
            from langchain_core.documents import Document

            loader = ApifyDatasetLoader(
                dataset_id="YOUR-DATASET-ID",
                dataset_mapping_function=lambda dataset_item: Document(
                    page_content=dataset_item["text"], metadata={"source": dataset_item["url"]}
                ),
            )
            documents = loader.load()
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    apify_client: ApifyClient
    """An instance of the ApifyClient class from the apify-client Python package."""
    dataset_id: str
    """The ID of the dataset on the Apify platform."""
    dataset_mapping_function: Callable[[dict], Document]
    """A custom function that takes a single dictionary (an Apify dataset item)
     and converts it to an instance of the Document class."""

    def __init__(
        self,
        dataset_id: str,
        dataset_mapping_function: Callable[[dict], Document],
        apify_api_token: str | None = None,
    ) -> None:
        """Initialize the loader with an Apify dataset ID and a mapping function.

        Args:
            dataset_id (str): The ID of the dataset on the Apify platform.
            dataset_mapping_function (Callable): A function that takes a single
                dictionary (an Apify dataset item) and converts it to an instance
                of the Document class.
            apify_api_token (str): Apify API token.
        """
        super().__init__(
            dataset_id=dataset_id,
            dataset_mapping_function=dataset_mapping_function,
            apify_api_token=apify_api_token,
        )

    @model_validator(mode='before')
    @classmethod
    def validate_environment(cls, values: dict) -> Any:  # noqa: ANN401
        """Validate environment.

        Args:
            values (dict): The values to validate.

        Returns:
            Any: The validated values.
        """
        apify_api_token = get_from_dict_or_env(values, 'apify_api_token', 'APIFY_API_TOKEN')

        client = create_apify_client(ApifyClient, apify_api_token)

        values['apify_client'] = client

        return values

    def load(self) -> list[Document]:
        """Load documents.

        Returns:
            list[Document]: A list of mapped Document objects.
        """
        dataset_items = self.apify_client.dataset(self.dataset_id).list_items(clean=True).items
        return list(map(self.dataset_mapping_function, dataset_items))

    def lazy_load(self) -> Iterator[Document]:
        """Lazily load documents.

        Yields:
            Document: A mapped Document object.
        """
        dataset_items = self.apify_client.dataset(self.dataset_id).iterate_items(
            clean=True,
        )
        for item in dataset_items:
            yield self.dataset_mapping_function(item)



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING, Any, Union

from apify_client import ApifyClient
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field, create_model

from langchain_apify.error_messages import ERROR_APIFY_TOKEN_ENV_VAR_NOT_SET
from langchain_apify.utils import (
    actor_id_to_tool_name,
    create_apify_client,
    get_actor_latest_build,
    prune_actor_input_schema,
)

from .const import MAX_DESCRIPTION_LEN

if TYPE_CHECKING:
    from langchain_core.callbacks import (
        CallbackManagerForToolRun,
    )


class ApifyActorsTool(BaseTool):  # type: ignore[override, override]
    """Tool that runs Apify Actors.

    To use, you should have the environment variable `APIFY_API_TOKEN` set
    with your API key, or pass `apify_api_token`
    as a named parameter to the constructor.

    For details, see https://docs.apify.com/platform/integrations/langchain

    Example:
        .. code-block:: python

            from langchain_apify import ApifyActorsTool
            from langgraph.prebuilt import create_react_agent

            tool = ApifyActorsTool(actor_id="apify/rag-web-browser")
            # Use the tool directly to call the Apify Actor
            result = tool.invoke(
                input={"run_input": {"query": "what is Apify?", "maxResults": 3}}
            )

            # Use the tool with an agent
            tools = [tool]
            agent = create_react_agent(model, tools)

            for chunk in agent.stream(
                {"messages": [("human", "search for what is Apify?")]},
                stream_mode="values"
            ):
                chunk["messages"][-1].pretty_print()
    """

    def __init__(
        self,
        actor_id: str,
        apify_api_token: str | None = None,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize the tool with an Apify Actor.

        Args:
            actor_id (str): Actor name from Apify store to run.
            apify_api_token (Optional[str]): Apify API token.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Raises:
            ValueError: If the `APIFY_API_TOKEN` environment variable is not set
        """
        apify_api_token = apify_api_token or os.getenv('APIFY_API_TOKEN')
        if not apify_api_token:
            msg = ERROR_APIFY_TOKEN_ENV_VAR_NOT_SET
            raise ValueError(msg)

        apify_client = create_apify_client(ApifyClient, apify_api_token)

        kwargs.update(
            {
                'name': actor_id_to_tool_name(actor_id),
                'description': self._create_description(apify_client, actor_id),
                'args_schema': self._build_tool_args_schema_model(
                    apify_client,
                    actor_id,
                ),
            },
        )

        super().__init__(*args, **kwargs)

        self._apify_client = apify_client
        self._actor_id = actor_id

    def _run(
        self,
        run_input: str | dict,
        _: CallbackManagerForToolRun | None = None,
    ) -> list[dict]:
        """Use the Apify Actor.

        Args:
            run_input (Union[str, dict]): JSON input for the Actor.

        Returns:
            list[dict]: The output dataset.
        """
        input_dict = json.loads(run_input) if isinstance(run_input, str) else run_input
        # retrieve if nested, just in case
        input_dict = input_dict.get('run_input', input_dict)
        return self._run_actor(input_dict)

    @staticmethod
    def _create_description(apify_client: ApifyClient, actor_id: str) -> str:
        """Create a description for the tool.

        Args:
            apify_client (ApifyClient): Apify client instance.
            actor_id (str): Actor name from Apify store to run.

        Returns:
            str: The description.
        """
        build = get_actor_latest_build(apify_client, actor_id)
        actor_description = build.get('actorDefinition', {}).get('description', '')
        if len(actor_description) > MAX_DESCRIPTION_LEN:
            actor_description = actor_description[:MAX_DESCRIPTION_LEN] + '...(TRUNCATED, TOO LONG)'
        return actor_description

    @staticmethod
    def _build_tool_args_schema_model(
        apify_client: ApifyClient,
        actor_id: str,
    ) -> type[BaseModel]:
        """Build a tool class for an agent that runs the Apify Actor.

        Args:
            apify_client (ApifyClient): Apify client instance.
            actor_id (str): Actor name from Apify store to run.

        Returns:
            type[BaseModel]: The tool input model class for the Apify Actor.

        Raises:
            ValueError: If the input schema is not found in the Actor build.
        """
        build = get_actor_latest_build(apify_client, actor_id)
        if not (actor_input := build.get('actorDefinition', {}).get('input')):
            msg = f'Input schema not found in the Actor build for Actor: {actor_id}'
            raise ValueError(msg)

        properties, required = prune_actor_input_schema(actor_input)
        properties = {'run_input': properties}

        description = (
            "JSON encoded as a string with input schema "
            "(STRICTLY FOLLOW JSON FORMAT AND SCHEMA):\n\n"
            f"{json.dumps(properties, separators=(',', ':'))}"
            "\nIF THE TOOL INPUT SCHEMA SUPPORTS IT LIMIT THE NUMBER OF RESULTS"
        )
        if required:
            description += '\n\nRequired fields:\n' + '\n'.join(required)

        return create_model(
            'ApifyActorsToolInput',
            run_input=(Union[str, dict], Field(..., description=description)),
        )

    def _run_actor(self, run_input: dict) -> list[dict]:
        """Run an Apify Actor and return the output dataset.

        Args:
            run_input: dict, JSON input for the Actor

        Returns:
            list[dict]: The output dataset

        Raises:
            ValueError: If the Actor was not started properly or the Run ID was not found in the run details
        """
        if (details := self._apify_client.actor(actor_id=self._actor_id).call(run_input=run_input)) is None:
            msg = f'Actor: {self._actor_id} was not started properly and details about the run were not returned'
            raise ValueError(msg)
        if (run_id := details.get('id')) is None:
            msg = f'Run ID not found in the run details for Actor: {self._actor_id}'
            raise ValueError(msg)
        run = self._apify_client.run(run_id=run_id)

        return run.dataset().list_items(clean=True).items



ERROR_APIFY_TOKEN_ENV_VAR_NOT_SET = (
    'APIFY_API_TOKEN environment variable is not set.'
    ' Please set it to your Apify API token by using `os.environ["APIFY_API_TOKEN"] = "YOUR_APIFY_API_TOKEN"'
    ' in your code or pass it as environment variable.'
    ' To pass it as environment variable, you can use the following command:'
    ' `APIFY_API_TOKEN="YOUR_APIFY_API_TOKEN" python your_script.py`'
)



from __future__ import annotations

import string
from typing import TypeVar

import requests
from apify_client import ApifyClientAsync
from apify_client.client import ApifyClient

from langchain_apify.const import MAX_DESCRIPTION_LEN, REQUESTS_TIMEOUT_SECS

APIFY_API_ENDPOINT_GET_DEFAULT_BUILD = 'https://api.apify.com/v2/acts/{actor_id}/builds/default'


def prune_actor_input_schema(
    input_schema: dict,
    max_description_len: int = MAX_DESCRIPTION_LEN,
) -> tuple[dict, list[str]]:
    """Get the input schema from the Actor build.

    Trim the description to 250 characters.

    Args:
        input_schema (dict): The input schema from the Actor build.
        max_description_len (int): The maximum length of the description.

    Returns:
        tuple[dict, list[str]]: A tuple containing the pruned properties
            and required fields.
    """
    properties = input_schema.get('properties', {})
    required = input_schema.get('required', [])

    properties_out: dict = {}
    for item, meta in properties.items():
        properties_out[item] = {}
        if desc := meta.get('description'):
            properties_out[item]['description'] = (
                desc[:max_description_len] + '...' if len(desc) > max_description_len else desc
            )
        for key_name in ('type', 'default', 'prefill', 'enum'):
            if value := meta.get(key_name):
                properties_out[item][key_name] = value

    return properties_out, required


T = TypeVar('T', ApifyClient, ApifyClientAsync)


def create_apify_client(client_cls: type[T], token: str) -> T:
    """Create an Apify client instance with a custom user-agent.

    Args:
        client_cls (ApifyClient | ApifyClientAsync): ApifyClient or ApifyClientAsync class.
        token (str): API token.

    Returns:
        T: ApifyClient or ApifyClientAsync instance.

    Raises:
        ValueError: If the API token is not provided.
    """
    if not token:
        msg = 'API token is required to create an Apify client.'
        raise ValueError(msg)
    client = client_cls(token)
    http_client_attr = 'httpx_async_client' if isinstance(client, ApifyClientAsync) else 'httpx_client'
    if http_client := getattr(client.http_client, http_client_attr):
        http_client.headers['user-agent'] += '; Origin/langchain'
    return client


def actor_id_to_tool_name(actor_id: str) -> str:
    """Turn actor_id into a valid tool name.

    Tool name must only contain letters, numbers, underscores, dashes,
    and cannot contain spaces.

    Args:
        actor_id (str): Actor ID from Apify store.

    Returns:
        str: A valid tool name.
    """
    valid_chars = string.ascii_letters + string.digits + '_-'
    return 'apify_actor_' + ''.join(char if char in valid_chars else '_' for char in actor_id)


def get_actor_latest_build(apify_client: ApifyClient, actor_id: str) -> dict:
    """Get the latest build of an Actor from the default build tag.

    Args:
        apify_client (ApifyClient): An instance of the ApifyClient class.
        actor_id (str): Actor name from Apify store to run.

    Returns:
        dict: The latest build of the Actor.

    Raises:
        ValueError: If the Actor is not found or the build data is not found.
        TypeError: If the build is not a dictionary.
    """
    if not (actor := apify_client.actor(actor_id).get()):
        msg = f'Actor {actor_id} not found.'
        raise ValueError(msg)

    if not (actor_obj_id := actor.get('id')):
        msg = f'Failed to get the Actor object ID for {actor_id}.'
        raise ValueError(msg)

    url = APIFY_API_ENDPOINT_GET_DEFAULT_BUILD.format(actor_id=actor_obj_id)
    response = requests.request('GET', url, timeout=REQUESTS_TIMEOUT_SECS)

    build = response.json()
    if not isinstance(build, dict):
        msg = f'Failed to get the latest build of the Actor {actor_id}.'
        raise TypeError(msg)

    if (data := build.get('data')) is None:
        msg = f'Failed to get the latest build data of the Actor {actor_id}.'
        raise ValueError(msg)

    return data



from importlib import metadata

from langchain_apify.document_loaders import ApifyDatasetLoader
from langchain_apify.tools import ApifyActorsTool
from langchain_apify.wrappers import ApifyWrapper

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ''
del metadata  # optional, avoids polluting the results of dir(__package__)

__all__ = [
    'ApifyActorsTool',
    'ApifyDatasetLoader',
    'ApifyWrapper',
    '__version__',
]
